import sys
import traceback
import time

import aergo.herapy as herapy


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    print("------ Payload -----------")
    """
-- 'payload' is compiled by aergoluac from
function setItem(key, value)
    system.print('setItem: key='..key..', value='..value)
    system.setItem(key, value)
end

function getItem(key)
    system.print('getItem: key='..key)
    return system.getItem(key)
end

abi.register(setItem, getItem)
    """
    payload_str = "246oyQ629vXweGRbeS8r29poBjGsiqJL23BBaYh8TaSgvGhMAXHrk9vUXetphe6m9xCPrRDpQH3fHap4SZUpNWartsHTnJxQXp2zs5mMwK5cwiCgSCePwefA1U4fwyt94Q249MxQ2evTKc7vcUVjjn6AnKCqMJTGCgVtnLrpeVMiVEcYBszJtprA3YQmgA8vsnfcw5ocw4149wDhV8drF4zviYJSfhGP5TRj7NhLLBstiktq2DzPW8yCrVu2fEnyzfJc1yY9ENuu2K7LyxC7arqGTafZEzsQrrfNo63PXfsSq6oSW7Ev89Y3cFfRjpQfr8mHZPmfCRykUF7aepvkH64r6S7Sg3uUpgrZCnDKuDRuBCizBHUxJDUCua3tJNA2dKFtx9dpSTfPFkdvPoBFsTjegu6DwHgDop5zZMPV5S4Q5oi5vB1UsvSoX81cMAfp38kJe45ko9rLZKk83zpSfgzMLnjDZWccmjtNdnHFhUp9KZG3JRZnr7QRmAHwvTYLd83AS9uCVJhwg5oGhVNgMeTnN2f1xZ9Gd9C5NfJTvwYHPVysD7HRwyXVTwYgQhjxwTrxkGcz54gJzF2eTDqWnEQKZN61322SZG"
    payload = herapy.utils.decode_address(payload_str)
    print(''.join('{:d} '.format(x) for x in payload))

    byte_code_len = int.from_bytes(payload[:4], byteorder='little')
    print("payload: byte code length = ", byte_code_len)

    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Set Sender Account -----------")
        sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
        sender_account = aergo.new_account(private_key=sender_private_key)
        print("  > Sender Address: {}".format(sender_account.address))
        print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Set Receiver Account -----------")
        receiver_address = "AmNHbk46L5ZaFH942mxDrunhUb34S8xRd7ygNnqaW5nqJDt5ugKD"
        print("  > Receiver Address: {}".format(receiver_address))

        print("------ Deploy SC -----------")
        tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234)
        print("  > TX: {}".format(tx.tx_hash))
        print("{}".format(herapy.utils.convert_tx_to_json(tx)))
        if result.status != herapy.CommitStatus.TX_OK:
            eprint("    > ERROR[{0}]: {1}".format(result.status, result.detail))
            aergo.disconnect()
            return
        else:
            print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
            print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

        time.sleep(3)

        print("------ Check deployment of SC -----------")
        print("  > TX: {}".format(tx.tx_hash))
        result = aergo.get_tx_result(tx.tx_hash)
        if result.status != herapy.TxResultStatus.CREATED:
            eprint("  > ERROR[{0}]:{1}: {2}".format(
                result.contract_address, result.status, result.detail))
            aergo.disconnect()
            return

        sc_address = result.contract_address
        print("  > SC Address: {}".format(sc_address))

        print("------ Batch Call SC -----------")
        nonce = aergo.account.nonce + 1
        sc_txs = [
            aergo.new_call_sc_tx(sc_address, "setItem", args=["key1", "value1"], nonce=nonce),
            aergo.new_call_sc_tx(sc_address, "setItem", args=["key2", "value2"], nonce=nonce+1),
            aergo.new_call_sc_tx(sc_address, "setItem", args=["key3", "value3"], nonce=nonce+2),
            aergo.new_call_sc_tx(sc_address, "setItem", args=["key4", "value4"], nonce=nonce+3),
            aergo.new_call_sc_tx(sc_address, "setItem", args=["key5", "value5"], nonce=nonce+4),
        ]
        aergo.batch_call_sc(sc_txs)

        time.sleep(3)

        print("------ Check result of Batch Call SC -----------")
        for i, tx in enumerate(sc_txs):
            print("  > TX[{0}] Result: {1}".format(i, str(tx.tx_hash)))
            result = aergo.get_tx_result(tx.tx_hash)
            if result.status != herapy.TxResultStatus.SUCCESS:
                eprint("  > ERROR[{0}]:{1}: {2}".format(
                    result.contract_address, result.status, result.detail))
                aergo.disconnect()
                return

            sc_address = result.contract_address
            print("  > SC Address: {}".format(sc_address))

        print("------ Query SC -----------")
        result = aergo.query_sc(sc_address, "getItem", args=["key1"])
        print(result)
        result = aergo.query_sc(sc_address, "getItem", args=["key2"])
        print(result)
        result = aergo.query_sc(sc_address, "getItem", args=["key3"])
        print(result)
        result = aergo.query_sc(sc_address, "getItem", args=["key4"])
        print(result)
        result = aergo.query_sc(sc_address, "getItem", args=["key5"])
        print(result)

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



import pytest

import aergo.herapy as herapy


DEFAULT_AERGO_CONFIG = """datadir = "${AERGO_HOME}/data"
dbtype = "badgerdb"
enableprofile = false
profileport = 6060
enabletestmode = false
usetestnet = false
personal = true
authdir = "${AERGO_HOME}/auth"

[rpc]
netserviceaddr = "127.0.0.1"
netserviceport = 7845
netservicetrace = false
nstls = false
nscert = ""
nskey = ""
nscacert = ""
nsallowcors = false

[p2p]
netprotocoladdr = ""
netprotocolport = 7846
npbindaddr = ""
npbindport = -1
nptls = false
npcert = ""
npkey = ""
npaddpeers = []
npdiscoverpeers = true
npmaxpeers = 100
nppeerpool = 100
npexposeself = true
npusepolaris = true
npaddpolarises = []
logfullpeerid = false

[polaris]
allowprivate = false
genesisfile = ""

[blockchain]
maxblocksize = 1048576
coinbaseaccount = ""
maxanchorcount = 20
forceresetheight = 0
zerofee = true
verifyonly = false
statetrace = 0

[mempool]
showmetrics = false
enablefadeout = false
fadeoutperiod = 12
dumpfilepath = "${AERGO_HOME}/mempool.dump"

[consensus]
enablebp = false
blockinterval = 1

[monitor]
protocol = ""
endpoint = ""

[account]
unlocktimeout = 60

[auth]
enablelocalconf = false
"""


def test_default():
    aergo_conf = herapy.AergoConfig()
    generate_toml = herapy.utils.convert_aergo_conf_to_toml(aergo_conf)
    #print(generate_toml)
    assert DEFAULT_AERGO_CONFIG == generate_toml
    with pytest.raises(KeyError):
        aergo_conf.p2p_nphiddenpeers
    with pytest.raises(KeyError):
        aergo_conf.blockchain_verifiercount
    with pytest.raises(KeyError):
        aergo_conf.mempool_verifiers


TEST_AERGO_CONFIG = """
# aergo TOML Configuration File (https://github.com/toml-lang/toml)
# base configurations
datadir = "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/data"
enableprofile = false
profileport = 6060
#enablerest = false
enabletestmode = true

[rpc]
netserviceaddr = "127.0.0.1"
netserviceport = 7845
nstls = false
nscert = ""
nskey = ""
nsallowcors = false

#[rest]
#restport = "8080"

[p2p]
netprotocoladdr = "127.0.0.1"
netprotocolport = 7846
nptls = false
npcert = ""
npkey = ""
npaddpeers = [
]
npmaxpeers = "100"
nppeerpool = "100"

[blockchain]
# blockchain configurations
maxblocksize = 1048576

[mempool]
showmetrics = false
dumpfilepath = "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/mempool.dump"

[consensus]
enablebp = true
blockinterval = 3
"""


def test_success():
    aergo_conf = herapy.utils.convert_toml_to_aergo_conf(TEST_AERGO_CONFIG)
    with pytest.raises(KeyError):
        aergo_conf.add_conf("test", "test")
    with pytest.raises(KeyError):
        aergo_conf.add_conf("test", "test", "test")
    with pytest.raises(KeyError):
        aergo_conf.add_conf("test", "test", "rpc")
    # check base config
    assert aergo_conf.datadir == "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/data"
    assert aergo_conf.dbtype == herapy.AERGO_DEFAULT_CONF['dbtype']
    assert aergo_conf.enableprofile is False
    assert aergo_conf.profileport == 6060
    with pytest.raises(KeyError):
        herapy.AERGO_DEFAULT_CONF['enablerest']
    assert aergo_conf.enabletestmode is True
    assert aergo_conf.personal is herapy.AERGO_DEFAULT_CONF['personal']
    assert aergo_conf.authdir is herapy.AERGO_DEFAULT_CONF['authdir']
    # check rcp config
    assert aergo_conf.rpc_netserviceaddr == "127.0.0.1"
    assert aergo_conf.rpc['netserviceaddr'] == aergo_conf.rpc_netserviceaddr
    assert aergo_conf.rpc_netserviceport == 7845
    assert aergo_conf.rpc['netserviceport'] == aergo_conf.rpc_netserviceport
    assert aergo_conf.rpc_netservicetrace is False
    assert aergo_conf.rpc['netservicetrace'] == aergo_conf.rpc_netservicetrace
    assert aergo_conf.rpc_nstls is False
    assert aergo_conf.rpc['nstls'] == aergo_conf.rpc_nstls
    assert aergo_conf.rpc_nscert == ""
    assert aergo_conf.rpc_nskey == ""
    assert aergo_conf.rpc_nsallowcors is False
    # check rest config
    #assert aergo_conf.rest_restport == 8080
    # check p2p config
    assert aergo_conf.p2p_netprotocoladdr != herapy.AERGO_DEFAULT_CONF['p2p']['netprotocoladdr']
    assert aergo_conf.p2p_netprotocoladdr == "127.0.0.1"
    assert aergo_conf.p2p['netprotocoladdr'] == aergo_conf.p2p_netprotocoladdr
    assert aergo_conf.p2p_netprotocolport == 7846
    assert aergo_conf.p2p['netprotocolport'] == aergo_conf.p2p_netprotocolport
    assert aergo_conf.p2p_npbindaddr == herapy.AERGO_DEFAULT_CONF['p2p']['npbindaddr']
    assert aergo_conf.p2p['npbindaddr'] == aergo_conf.p2p_npbindaddr
    assert aergo_conf.p2p_npbindport == herapy.AERGO_DEFAULT_CONF['p2p']['npbindport']
    assert aergo_conf.p2p['npbindport'] == aergo_conf.p2p_npbindport
    assert aergo_conf.p2p_nptls is False
    assert aergo_conf.p2p['nptls'] == aergo_conf.p2p_nptls
    assert aergo_conf.p2p_npcert == ""
    assert aergo_conf.p2p['npcert'] == aergo_conf.p2p_npcert
    assert aergo_conf.p2p_npkey == ""
    assert aergo_conf.p2p['npkey'] == aergo_conf.p2p_npkey
    assert isinstance(aergo_conf.p2p_npaddpeers, list)
    assert 0 == len(aergo_conf.p2p_npaddpeers)
    assert aergo_conf.p2p_npmaxpeers == 100
    assert int(aergo_conf.p2p['npmaxpeers']) == aergo_conf.p2p_npmaxpeers
    assert aergo_conf.p2p_nppeerpool == 100
    assert int(aergo_conf.p2p['nppeerpool']) == aergo_conf.p2p_nppeerpool
    # check blockchain config
    assert aergo_conf.blockchain_maxblocksize == 1048576
    assert aergo_conf.blockchain_coinbaseaccount == herapy.AERGO_DEFAULT_CONF['blockchain']['coinbaseaccount']
    assert aergo_conf.blockchain_maxanchorcount == herapy.AERGO_DEFAULT_CONF['blockchain']['maxanchorcount']
    #assert aergo_conf.blockchain_usefastsyncer is herapy.AERGO_DEFAULT_CONF['blockchain']['usefastsyncer']
    assert aergo_conf.blockchain_forceresetheight == herapy.AERGO_DEFAULT_CONF['blockchain']['forceresetheight']
    # check mempool config
    assert aergo_conf.mempool_showmetrics is False
    with pytest.raises(KeyError):
        herapy.AERGO_DEFAULT_CONF['mempool']['verifiers']
    with pytest.raises(KeyError):
        aergo_conf.mempool_verifiers
    assert aergo_conf.mempool_dumpfilepath == "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/mempool.dump"
    # check consensus config
    assert aergo_conf.consensus_enablebp is True
    assert aergo_conf.consensus_blockinterval == 3
    # check monitor config
    assert aergo_conf.monitor_protocol == herapy.AERGO_DEFAULT_CONF['monitor']['protocol']
    assert aergo_conf.monitor_endpoint == herapy.AERGO_DEFAULT_CONF['monitor']['endpoint']
    # check account config
    assert aergo_conf.account_unlocktimeout == herapy.AERGO_DEFAULT_CONF['account']['unlocktimeout']


def test_fail():
    aergo_conf = herapy.AergoConfig()
    # check datadir type
    aergo_conf.datadir = 'str'
    with pytest.raises(TypeError):
        aergo_conf.datadir = 1234
    with pytest.raises(TypeError):
        aergo_conf.datadir = True
    with pytest.raises(TypeError):
        aergo_conf.datadir = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.datadir = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.datadir = ['1234', 1234, ]
    # check dbtype type
    aergo_conf.dbtype = 'str'
    with pytest.raises(TypeError):
        aergo_conf.dbtype = 1234
    with pytest.raises(TypeError):
        aergo_conf.dbtype = True
    with pytest.raises(TypeError):
        aergo_conf.dbtype = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.dbtype = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.dbtype = ['1234', 1234, ]
    # check enableprofile type
    aergo_conf.enableprofile = True
    with pytest.raises(TypeError):
        aergo_conf.enableprofile = 1234
    with pytest.raises(TypeError):
        aergo_conf.enableprofile = '1234'
    with pytest.raises(TypeError):
        aergo_conf.enableprofile = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.enableprofile = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.enableprofile = ['1234', 1234, ]
    # check profileport type
    aergo_conf.profileport = 1234
    aergo_conf.profileport = False
    with pytest.raises(TypeError):
        aergo_conf.profileport = '1234'
    with pytest.raises(TypeError):
        aergo_conf.profileport = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.profileport = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.profileport = ['1234', 1234, ]
    # check enabletestmode type
    aergo_conf.enabletestmode = True
    with pytest.raises(TypeError):
        aergo_conf.enabletestmode = 1234
    with pytest.raises(TypeError):
        aergo_conf.enabletestmode = '1234'
    with pytest.raises(TypeError):
        aergo_conf.enabletestmode = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.enabletestmode = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.enabletestmode = ['1234', 1234, ]
    # check personal type
    aergo_conf.personal = True
    with pytest.raises(TypeError):
        aergo_conf.personal = 1234
    with pytest.raises(TypeError):
        aergo_conf.personal = '1234'
    with pytest.raises(TypeError):
        aergo_conf.personal = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.personal = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.personal = ['1234', 1234, ]
    # check authdir type
    aergo_conf.authdir = 'str'
    with pytest.raises(TypeError):
        aergo_conf.authdir = 1234
    with pytest.raises(TypeError):
        aergo_conf.authdir = True
    with pytest.raises(TypeError):
        aergo_conf.authdir = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.authdir = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.authdir = ['1234', 1234, ]
    # check usetestnet type
    aergo_conf.usetestnet = True
    with pytest.raises(TypeError):
        aergo_conf.usetestnet = 1234
    with pytest.raises(TypeError):
        aergo_conf.usetestnet = '1234'
    with pytest.raises(TypeError):
        aergo_conf.usetestnet = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.usetestnet = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.usetestnet = ['1234', 1234, ]
    # check rpc_netserviceaddr type
    aergo_conf.rpc_netserviceaddr = 'str'
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceaddr = 1234
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceaddr = True
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceaddr = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceaddr = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceaddr = ['1234', 1234, ]
    # check rpc_netserviceport type
    aergo_conf.rpc_netserviceport = 1234
    aergo_conf.rpc_netserviceport = False
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceport = '1234'
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceport = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceport = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.rpc_netserviceport = ['1234', 1234, ]
    # check rpc_netservicetrace type
    aergo_conf.rpc_netservicetrace = True
    with pytest.raises(TypeError):
        aergo_conf.rpc_netservicetrace = 1234
    with pytest.raises(TypeError):
        aergo_conf.rpc_netservicetrace = '1234'
    with pytest.raises(TypeError):
        aergo_conf.rpc_netservicetrace = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.rpc_netservicetrace = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.rpc_netservicetrace = ['1234', 1234, ]
    # check rpc_nstls type
    aergo_conf.rpc_nstls = True
    with pytest.raises(TypeError):
        aergo_conf.rpc_nstls = 1234
    with pytest.raises(TypeError):
        aergo_conf.rpc_nstls = '1234'
    with pytest.raises(TypeError):
        aergo_conf.rpc_nstls = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.rpc_nstls = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.rpc_nstls = ['1234', 1234, ]
    # check rpc_nscert type
    aergo_conf.rpc_nscert = 'str'
    with pytest.raises(TypeError):
        aergo_conf.rpc_nscert = 1234
    with pytest.raises(TypeError):
        aergo_conf.rpc_nscert = True
    with pytest.raises(TypeError):
        aergo_conf.rpc_nscert = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.rpc_nscert = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.rpc_nscert = ['1234', 1234, ]
    # check rpc_nskey type
    aergo_conf.rpc_nskey = 'str'
    with pytest.raises(TypeError):
        aergo_conf.rpc_nskey = 1234
    with pytest.raises(TypeError):
        aergo_conf.rpc_nskey = True
    with pytest.raises(TypeError):
        aergo_conf.rpc_nskey = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.rpc_nskey = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.rpc_nskey = ['1234', 1234, ]
    # check rpc_nsallowcors type
    aergo_conf.rpc_nsallowcors = True
    with pytest.raises(TypeError):
        aergo_conf.rpc_nsallowcors = 1234
    with pytest.raises(TypeError):
        aergo_conf.rpc_nsallowcors = '1234'
    with pytest.raises(TypeError):
        aergo_conf.rpc_nsallowcors = (1234, '1234')
    with pytest.raises(TypeError):
        aergo_conf.rpc_nsallowcors = {'1234'}
    with pytest.raises(TypeError):
        aergo_conf.rpc_nsallowcors = ['1234', 1234, ]



import sys
import traceback
import time

import aergo.herapy as herapy


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Set Sender Account -----------")
        sender_private_key = "eHoEcHnaxpGpgzknXjuwon8VFVrLkKHC4FckGuGkQ8depiDDfyUAWC3L"
        sender_account = aergo.new_account(private_key=sender_private_key)
        print("  > Sender Address: {}".format(sender_account.address))
        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Set Receiver Account -----------")
        receiver_address = "AmMPQqRJ4pd8bS9xdkw5pjExtLjaUXaYGB5kagzHu4A9ckKgnBV2"
        print("  > Receiver Address: {}".format(receiver_address))
        receiver_account = aergo.get_account(address=receiver_address)
        print("    > account state of Sender")
        print("      - balance        = {}".format(receiver_account.balance))
        print("      - nonce          = {}".format(receiver_account.nonce))
        print("      - code hash      = {}".format(receiver_account.code_hash))
        print("      - storage root   = {}".format(receiver_account.storage_root))

        print("------ Simple Send Tx -----------")
        simple_tx, result = aergo.send_payload(to_address=receiver_address,
                                               amount=10, payload=None,
                                               retry_nonce=3)
        print("  > simple TX[{}]".format(simple_tx.calculate_hash()))
        print("{}".format(str(simple_tx)))
        print("  > result: ", result)
        if result.status != herapy.CommitStatus.TX_OK:
            eprint("    > ERROR[{0}]: {1}".format(result.status, result.detail))
            aergo.disconnect()
            return
        else:
            print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))

        time.sleep(2)
        # Can check TX in mem pool or block after some sec later

        print("------ Check TX in Aergo -----------", flush=True)
        tx = aergo.get_tx(simple_tx.tx_hash)
        print(" TX from Aergo: ", str(tx))
        tx = tx.block.get_tx(tx.index_in_block)
        print(" TX in Block: ", str(tx))

        time.sleep(3)

        print("------ Check TX status -----------")
        print("  > TX: {}".format(str(simple_tx.tx_hash)))
        tx_result = aergo.get_tx_result(simple_tx.tx_hash)
        print("      result: ", tx_result)
        if tx_result.status != herapy.TxResultStatus.SUCCESS:
            eprint("  > ERROR[{0}]:{1}: {2}".format(
                tx_result.contract_address, tx_result.status, tx_result.detail))
            aergo.disconnect()
            return

        print("------ Get Sender Account Info -----------")
        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Create Tx -----------")
        tx = herapy.Transaction(from_address=bytes(sender_account.address),
                                nonce=sender_account.nonce + 1,
                                amount=10)
        tx.to_address = herapy.utils.decode_address(receiver_address)
        tx.chain_id = aergo.chain_id
        print("  > unsigned TX Hash: {}".format(tx.tx_hash))
        print("  > unsigned TX : {}".format(herapy.utils.convert_tx_to_formatted_json(tx)))

        print("------ Sign Tx -----------")
        tx.sign = sender_account.sign_msg_hash(tx.calculate_hash(including_sign=False))
        print("  > TX Signature: {}".format(tx.sign_str))
        print("  > TX Hash: {}".format(tx.tx_hash))
        print("  > TX : {}".format(herapy.utils.convert_tx_to_formatted_json(tx)))

        print("------ Send Tx -----------")
        tx, result = aergo.send_tx(tx)
        print("{}".format(herapy.utils.convert_tx_to_formatted_json(tx)))
        if result.status != herapy.CommitStatus.TX_OK:
            eprint("    > ERROR[{0}]: {1}".format(result.status, result.detail))
            aergo.disconnect()
            return
        else:
            print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))

        # wait to generate a block
        time.sleep(3)

        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: metric.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='metric.proto',
  package='types',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0cmetric.proto\x12\x05types\"2\n\x0eMetricsRequest\x12 \n\x05types\x18\x01 \x03(\x0e\x32\x11.types.MetricType\"+\n\x07Metrics\x12 \n\x05peers\x18\x01 \x03(\x0b\x32\x11.types.PeerMetric\"Z\n\nPeerMetric\x12\x0e\n\x06peerID\x18\x01 \x01(\x0c\x12\r\n\x05sumIn\x18\x02 \x01(\x03\x12\r\n\x05\x61vrIn\x18\x03 \x01(\x03\x12\x0e\n\x06sumOut\x18\x04 \x01(\x03\x12\x0e\n\x06\x61vrOut\x18\x05 \x01(\x03**\n\nMetricType\x12\x0b\n\x07NOTHING\x10\x00\x12\x0f\n\x0bP2P_NETWORK\x10\x01\x62\x06proto3')
)

_METRICTYPE = _descriptor.EnumDescriptor(
  name='MetricType',
  full_name='types.MetricType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NOTHING', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='P2P_NETWORK', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=212,
  serialized_end=254,
)
_sym_db.RegisterEnumDescriptor(_METRICTYPE)

MetricType = enum_type_wrapper.EnumTypeWrapper(_METRICTYPE)
NOTHING = 0
P2P_NETWORK = 1



_METRICSREQUEST = _descriptor.Descriptor(
  name='MetricsRequest',
  full_name='types.MetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='types', full_name='types.MetricsRequest.types', index=0,
      number=1, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23,
  serialized_end=73,
)


_METRICS = _descriptor.Descriptor(
  name='Metrics',
  full_name='types.Metrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peers', full_name='types.Metrics.peers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=118,
)


_PEERMETRIC = _descriptor.Descriptor(
  name='PeerMetric',
  full_name='types.PeerMetric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peerID', full_name='types.PeerMetric.peerID', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sumIn', full_name='types.PeerMetric.sumIn', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='avrIn', full_name='types.PeerMetric.avrIn', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sumOut', full_name='types.PeerMetric.sumOut', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='avrOut', full_name='types.PeerMetric.avrOut', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=210,
)

_METRICSREQUEST.fields_by_name['types'].enum_type = _METRICTYPE
_METRICS.fields_by_name['peers'].message_type = _PEERMETRIC
DESCRIPTOR.message_types_by_name['MetricsRequest'] = _METRICSREQUEST
DESCRIPTOR.message_types_by_name['Metrics'] = _METRICS
DESCRIPTOR.message_types_by_name['PeerMetric'] = _PEERMETRIC
DESCRIPTOR.enum_types_by_name['MetricType'] = _METRICTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MetricsRequest = _reflection.GeneratedProtocolMessageType('MetricsRequest', (_message.Message,), dict(
  DESCRIPTOR = _METRICSREQUEST,
  __module__ = 'metric_pb2'
  # @@protoc_insertion_point(class_scope:types.MetricsRequest)
  ))
_sym_db.RegisterMessage(MetricsRequest)

Metrics = _reflection.GeneratedProtocolMessageType('Metrics', (_message.Message,), dict(
  DESCRIPTOR = _METRICS,
  __module__ = 'metric_pb2'
  # @@protoc_insertion_point(class_scope:types.Metrics)
  ))
_sym_db.RegisterMessage(Metrics)

PeerMetric = _reflection.GeneratedProtocolMessageType('PeerMetric', (_message.Message,), dict(
  DESCRIPTOR = _PEERMETRIC,
  __module__ = 'metric_pb2'
  # @@protoc_insertion_point(class_scope:types.PeerMetric)
  ))
_sym_db.RegisterMessage(PeerMetric)


# @@protoc_insertion_point(module_scope)



import hashlib
import sys
import traceback
import time

import aergo.herapy as herapy
from aergo.herapy.obj.sc_state import SCStateVar


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    print("------ Payload -----------")
    """
        -- Define global variables.
        state.var {
            a = state.value(),
            b = state.array(3),
            c = state.map(),
        }

        function constructor(name)
            a:set("a_" .. name)
            b[1] = "b_" .. name
            c["init"] = "c_" .. name
            c[string.char(0x01)] = name
            c[fromhex('61')] = name
        end

        function fromhex(str)
            return (str:gsub('..', function (cc)
                return string.char(tonumber(cc, 16))
            end))
        end

        function set_name(key, name)
            if key == "a" then
                a:set(name)
            elseif key == "b" then
                b[2] = name
            else
                c['name'] = name
            end
        end

        function get_name(key)
            if key == "a" then
                return a:get()
            elseif key == "b" then
                return b[2]
            else
                return c['name']
            end
        end

        function say_hello()
            a:set("hello A")
            b[3] = "hello B"
            c["hello"] = "hello C"
        end

        abi.register(set_name, get_name, say_hello)
    """
    payload_str = "MyFxtNVdegMvf25CroFLsQs1H8mNXEdTwT7QTHYakg8gx8EJdivi87GJpLu9RUmEB7kECTMhcEeNYNRF7uW1rozSCBJameJZejrs2u39E57zebBCmTvp4331tAeVTQzr5B4PCt6xVd9UtU4kCBiL4Jfi8YJpu9jE3cJ5qpW39yZCoaSNMRWEsoDLePn1tBxnvvKTzP37o7Zges9gLoThS6G7ioH6cFQhSxV9fPBoPfE1b6zMoSi97tyaDDc1XjDJ4YCLLBhaVeN55ZY5ayjajBCoKXGTUp8yiQNkGx3Jy5YwYPZxnvtiYxzpoKZtYz3nA6QRgkC16ecFtGBM8GYzFhYoDQpse24yjQfqwH5SkQh6EHGzUaCzX7AahivvjkyAtdpS5kfbJqquYUoYHoqXXLcLTj5CZ9Wzck6mdq25o8sQmM8DBnQ3DX6QQyyY9agCcRHYkK5Zr52vDmigsXS1qXJfDvBmrUTsvXFMfPE3EZUV5PaYTNubS7AGjReWQpou5FCpkshoAdfoiLLHTa1yCZJdCbjv8gpPWJCeivFbUo6Y2Xp41HreLExnZHHpEeZvEgat8ArA24ardELWWiVeE9ZkW3VE7gReDUJNWStZBrASNje7N8GsHhdvdKSKGJ8LAFkF4MrDSH4hPpjZkK6PJBu2jCuX9VoMUcR3p6RJ6NZFf7JudiWEkkZx7kULbeN31rpAhm3bXgFnU8eiamQk4ddY2Qur6kfzjCTmdu9SUwi71b778LsmxsosqbpoFnbvUhBuZKrjuNA1uJdhSoErDe6RkpkP5bm6asszpNK7zBu5iUyn1VepnbCyG2zWdeK7UQufcLGQh8yFamY2FqKpBkxaXkSFyfSjE41VT2mgGFhkMdz4E5ddmPjSg4RGYw33V1tCq3sRtFRe34zR3HrRsr54yFLLMbS1tyzt6tMREAghhTgXaKt9akigkeunczpntdRcBFzNgGqBx2BHyBXaRPzUWWdouvCBjaUzeZuxKVhodU1uts6tHQJWjk4yvgSmDB1mooTaYTuHAWgCLrjUEMyTNZWkdsgKkxfZtV5Q8rbqJ8FGQvSCucXDjFyCy5XPf5GNxhvdj7qYuewMxMM428qF6cR8MBqCo9LX85iRPCKdSVksAedk1GWehVAzwqGMGgMAE21P1EzCdTf3khhkXNT5xxYHtYVGtFUoWPKqi5Xr1jc6h3yYVPjXuUvYPPCDYQU6mboSSAfr211XFsJS42WRF2xHVFRn3jT76kgqfNxbCfXeWrcRVJfLcwE9FACfXGbp7GdUUZbcYKjCCjJ68W6qznwbZuph9oCiPsM1Y4A4UaesSwoijVUHxwSgZc1uJWjBBzuV1APQGTW7Z6DhfoZKfKhtDabTQhrafoUB9ZCbCqqDdTSGBufYRJ6qsSqTfr15YEyKgzdy5juRrXADF17P3f4XCeerD5N8neUjsUfVa8BWFuQfDWeHskzGWtLQnVs2qfhqM8tUefuMBEp2vG3CgRe247WKDLPPHpPeeern7PWiJgf5JzRt9h3brCcQy5qp39KKznEdrVrS6iNMEBK5qjtPHSqFbdSkkYEbqRp6gdisSPGFS2jraXFi2x"
    payload = herapy.utils.decode_address(payload_str)

    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Set Sender Account -----------")
        sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
        sender_account = aergo.new_account(private_key=sender_private_key)
        print("  > Sender Address: {}".format(sender_account.address))
        print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Deploy SC -----------")
        tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234)
        print("  > TX: {}".format(tx.tx_hash))
        print("{}".format(herapy.utils.convert_tx_to_json(tx)))
        if result.status != herapy.CommitStatus.TX_OK:
            eprint("    > ERROR[{0}]: {1}".format(result.status, result.detail))
            aergo.disconnect()
            return
        else:
            print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
            print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

        time.sleep(3)

        print("------ Check deployment of SC -----------")
        print("  > TX: {}".format(tx.tx_hash))
        result = aergo.get_tx_result(tx.tx_hash)
        if result.status != herapy.TxResultStatus.CREATED:
            eprint("  > ERROR[{0}]:{1}: {2}".format(
                result.contract_address, result.status, result.detail))
            aergo.disconnect()
            return

        sc_address = result.contract_address
        print("  > SC Address: {}".format(sc_address))

        print("------ Query SC State Var with Proof -----------")
        best_block_hash, best_block_height = aergo.get_blockchain_status()
        block = aergo.get_block(best_block_hash)
        root = block.blocks_root_hash
        sc_state = aergo.query_sc_state(sc_address, ["_sv_a"], root=root)
        sc_root = sc_state.account.state_proof.state.storageRoot
        print("Variable Proof: {}".format(sc_state.var_proofs.verify_var_proof(
            sc_root, sc_state.var_proofs[0], hashlib.sha256(bytes("_sv_a", 'latin-1')).digest())))
        print("Variables Proof: {}".format(sc_state.var_proofs.verify_proof(root=sc_root)))
        print("Total Proof: {}".format(sc_state.verify_proof(root)))

        storage_keys = ["_sv_a", "_sv_b-1", "_sv_c-init", "_sv_c-\x01", "_sv_c-a"]
        sc_state = aergo.query_sc_state(sc_address, storage_keys, root=root)
        print("Proof: {}".format(sc_state.verify_proof(root)))
        for i, vp in enumerate(sc_state.var_proofs):
            print("[{0}] Variable Proof: {1}".format(i, sc_state.var_proofs.verify_var_proof(
                sc_root, vp, hashlib.sha256(bytes(storage_keys[i], 'latin-1')).digest())))
            print("  var_proof.key          = {}".format(vp.key))
            print("  var_proof.value        = {}".format(vp.value))
            print("  var_proof.inclusion    = {}".format(vp.inclusion))
            print("  var_proof.proof_key    = {}".format(vp.proofKey))
            print("  var_proof.proof_value  = {}".format(vp.proofVal))
            print("  var_proof.bitmap       = {}".format(vp.bitmap))
            print("  var_proof.height       = {}".format(vp.height))

        sc_var_a = SCStateVar(var_name="a")
        sc_var_b = SCStateVar(var_name="b", array_index=1)
        sc_var_c = SCStateVar(var_name="c", map_key='init')
        sc_state = aergo.query_sc_state(sc_address, [sc_var_a, sc_var_b, sc_var_c], root=root, compressed=False)
        print("Proof: {}".format(sc_state.verify_proof(root)))

        sc_state = aergo.query_sc_state(sc_address, ['not', 'included', 'var'], root=root)
        for i, vp in enumerate(sc_state.var_proofs):
            print("var[{}]".format(i))
            print("  key: {}".format(vp.key))
            print("  inclusion? {}".format(vp.inclusion))
        print("not inclusion? Proof = {}".format(sc_state.verify_proof(root)))

        sc_state = aergo.query_sc_state(sc_address, ['not', 'included', 'var'], root=root, compressed=False)
        for i, vp in enumerate(sc_state.var_proofs):
            print("var[{}]".format(i))
            print("  key: {}".format(vp.key))
            print("  inclusion? {}".format(vp.inclusion))
        print("not inclusion? Proof = {}".format(sc_state.verify_proof(root)))

        address = "AmMejL8z3wW2doksBzzMiWM2xTb6WtZniLkLyxwqWKiLJKK8Yvqd"
        sc_state = aergo.query_sc_state(address, ["_sv_a"], root=root)
        sc_root = sc_state.account.state_proof.state.storageRoot
        print("Number of Variable Proofs = {}".format(len(sc_state.var_proofs)))
        print("Variables Proof: {}".format(sc_state.var_proofs.verify_proof(root=sc_root)))
        print("Total Proof: {}".format(sc_state.verify_proof(root)))

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

from .stream import Stream
from .block import Block


class BlockStream(Stream):
    def __init__(self, block_stream):
        super().__init__(block_stream)

    def next(self):
        return Block(grpc_block=next(self._grpc_stream))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: blockchain.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='blockchain.proto',
  package='types',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x10\x62lockchain.proto\x12\x05types\"Y\n\x05\x42lock\x12\x0c\n\x04hash\x18\x01 \x01(\x0c\x12\"\n\x06header\x18\x02 \x01(\x0b\x32\x12.types.BlockHeader\x12\x1e\n\x04\x62ody\x18\x03 \x01(\x0b\x32\x10.types.BlockBody\"\xe9\x01\n\x0b\x42lockHeader\x12\x0f\n\x07\x63hainID\x18\x01 \x01(\x0c\x12\x15\n\rprevBlockHash\x18\x02 \x01(\x0c\x12\x0f\n\x07\x62lockNo\x18\x03 \x01(\x04\x12\x11\n\ttimestamp\x18\x04 \x01(\x03\x12\x16\n\x0e\x62locksRootHash\x18\x05 \x01(\x0c\x12\x13\n\x0btxsRootHash\x18\x06 \x01(\x0c\x12\x18\n\x10receiptsRootHash\x18\x07 \x01(\x0c\x12\x10\n\x08\x63onfirms\x18\x08 \x01(\x04\x12\x0e\n\x06pubKey\x18\t \x01(\x0c\x12\x17\n\x0f\x63oinbaseAccount\x18\n \x01(\x0c\x12\x0c\n\x04sign\x18\x0b \x01(\x0c\"#\n\tBlockBody\x12\x16\n\x03txs\x18\x01 \x03(\x0b\x32\t.types.Tx\" \n\x06TxList\x12\x16\n\x03txs\x18\x01 \x03(\x0b\x32\t.types.Tx\"/\n\x02Tx\x12\x0c\n\x04hash\x18\x01 \x01(\x0c\x12\x1b\n\x04\x62ody\x18\x02 \x01(\x0b\x32\r.types.TxBody\"\xc0\x01\n\x06TxBody\x12\r\n\x05nonce\x18\x01 \x01(\x04\x12\x0f\n\x07\x61\x63\x63ount\x18\x02 \x01(\x0c\x12\x11\n\trecipient\x18\x03 \x01(\x0c\x12\x0e\n\x06\x61mount\x18\x04 \x01(\x0c\x12\x0f\n\x07payload\x18\x05 \x01(\x0c\x12\x10\n\x08gasLimit\x18\x06 \x01(\x04\x12\x10\n\x08gasPrice\x18\x07 \x01(\x0c\x12\x1b\n\x04type\x18\x08 \x01(\x0e\x32\r.types.TxType\x12\x13\n\x0b\x63hainIdHash\x18\t \x01(\x0c\x12\x0c\n\x04sign\x18\n \x01(\x0c\"\'\n\x05TxIdx\x12\x11\n\tblockHash\x18\x01 \x01(\x0c\x12\x0b\n\x03idx\x18\x02 \x01(\x05\"?\n\tTxInBlock\x12\x1b\n\x05txIdx\x18\x01 \x01(\x0b\x32\x0c.types.TxIdx\x12\x15\n\x02tx\x18\x02 \x01(\x0b\x32\t.types.Tx\"h\n\x05State\x12\r\n\x05nonce\x18\x01 \x01(\x04\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x0c\x12\x10\n\x08\x63odeHash\x18\x03 \x01(\x0c\x12\x13\n\x0bstorageRoot\x18\x04 \x01(\x0c\x12\x18\n\x10sqlRecoveryPoint\x18\x05 \x01(\x04\"\xa2\x01\n\x0c\x41\x63\x63ountProof\x12\x1b\n\x05state\x18\x01 \x01(\x0b\x32\x0c.types.State\x12\x11\n\tinclusion\x18\x02 \x01(\x08\x12\x0b\n\x03key\x18\x03 \x01(\x0c\x12\x10\n\x08proofKey\x18\x04 \x01(\x0c\x12\x10\n\x08proofVal\x18\x05 \x01(\x0c\x12\x0e\n\x06\x62itmap\x18\x06 \x01(\x0c\x12\x0e\n\x06height\x18\x07 \x01(\r\x12\x11\n\tauditPath\x18\x08 \x03(\x0c\"\x9e\x01\n\x10\x43ontractVarProof\x12\r\n\x05value\x18\x01 \x01(\x0c\x12\x11\n\tinclusion\x18\x02 \x01(\x08\x12\x10\n\x08proofKey\x18\x04 \x01(\x0c\x12\x10\n\x08proofVal\x18\x05 \x01(\x0c\x12\x0e\n\x06\x62itmap\x18\x06 \x01(\x0c\x12\x0e\n\x06height\x18\x07 \x01(\r\x12\x11\n\tauditPath\x18\x08 \x03(\x0c\x12\x0b\n\x03key\x18\t \x01(\x0cJ\x04\x08\x03\x10\x04\"i\n\x0fStateQueryProof\x12*\n\rcontractProof\x18\x01 \x01(\x0b\x32\x13.types.AccountProof\x12*\n\tvarProofs\x18\x02 \x03(\x0b\x32\x17.types.ContractVarProof\"\xf7\x01\n\x07Receipt\x12\x17\n\x0f\x63ontractAddress\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x0b\n\x03ret\x18\x03 \x01(\t\x12\x0e\n\x06txHash\x18\x04 \x01(\x0c\x12\x0f\n\x07\x66\x65\x65Used\x18\x05 \x01(\x0c\x12\x19\n\x11\x63umulativeFeeUsed\x18\x06 \x01(\x0c\x12\r\n\x05\x62loom\x18\x07 \x01(\x0c\x12\x1c\n\x06\x65vents\x18\x08 \x03(\x0b\x32\x0c.types.Event\x12\x0f\n\x07\x62lockNo\x18\t \x01(\x04\x12\x11\n\tblockHash\x18\n \x01(\x0c\x12\x0f\n\x07txIndex\x18\x0b \x01(\x05\x12\x0c\n\x04\x66rom\x18\x0c \x01(\x0c\x12\n\n\x02to\x18\r \x01(\x0c\"\x9c\x01\n\x05\x45vent\x12\x17\n\x0f\x63ontractAddress\x18\x01 \x01(\x0c\x12\x11\n\teventName\x18\x02 \x01(\t\x12\x10\n\x08jsonArgs\x18\x03 \x01(\t\x12\x10\n\x08\x65ventIdx\x18\x04 \x01(\x05\x12\x0e\n\x06txHash\x18\x05 \x01(\x0c\x12\x11\n\tblockHash\x18\x06 \x01(\x0c\x12\x0f\n\x07\x62lockNo\x18\x07 \x01(\x04\x12\x0f\n\x07txIndex\x18\x08 \x01(\x05\"\x1a\n\nFnArgument\x12\x0c\n\x04name\x18\x01 \x01(\t\"]\n\x08\x46unction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\targuments\x18\x02 \x03(\x0b\x32\x11.types.FnArgument\x12\x0f\n\x07payable\x18\x03 \x01(\x08\x12\x0c\n\x04view\x18\x04 \x01(\x08\"3\n\x08StateVar\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x0b\n\x03len\x18\x03 \x01(\x05\"v\n\x03\x41\x42I\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x10\n\x08language\x18\x02 \x01(\t\x12\"\n\tfunctions\x18\x03 \x03(\x0b\x32\x0f.types.Function\x12(\n\x0fstate_variables\x18\x04 \x03(\x0b\x32\x0f.types.StateVar\"3\n\x05Query\x12\x17\n\x0f\x63ontractAddress\x18\x01 \x01(\x0c\x12\x11\n\tqueryinfo\x18\x02 \x01(\x0c\"b\n\nStateQuery\x12\x17\n\x0f\x63ontractAddress\x18\x01 \x01(\x0c\x12\x0c\n\x04root\x18\x03 \x01(\x0c\x12\x12\n\ncompressed\x18\x04 \x01(\x08\x12\x13\n\x0bstorageKeys\x18\x05 \x03(\x0cJ\x04\x08\x02\x10\x03\"\x95\x01\n\nFilterInfo\x12\x17\n\x0f\x63ontractAddress\x18\x01 \x01(\x0c\x12\x11\n\teventName\x18\x02 \x01(\t\x12\x11\n\tblockfrom\x18\x03 \x01(\x04\x12\x0f\n\x07\x62lockto\x18\x04 \x01(\x04\x12\x0c\n\x04\x64\x65sc\x18\x05 \x01(\x08\x12\x11\n\targFilter\x18\x06 \x01(\x0c\x12\x16\n\x0erecentBlockCnt\x18\x07 \x01(\x05*2\n\x06TxType\x12\n\n\x06NORMAL\x10\x00\x12\x0e\n\nGOVERNANCE\x10\x01\x12\x0c\n\x08REDEPLOY\x10\x02\x62\x06proto3')
)

_TXTYPE = _descriptor.EnumDescriptor(
  name='TxType',
  full_name='types.TxType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NORMAL', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GOVERNANCE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REDEPLOY', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2324,
  serialized_end=2374,
)
_sym_db.RegisterEnumDescriptor(_TXTYPE)

TxType = enum_type_wrapper.EnumTypeWrapper(_TXTYPE)
NORMAL = 0
GOVERNANCE = 1
REDEPLOY = 2



_BLOCK = _descriptor.Descriptor(
  name='Block',
  full_name='types.Block',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash', full_name='types.Block.hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='header', full_name='types.Block.header', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='body', full_name='types.Block.body', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=27,
  serialized_end=116,
)


_BLOCKHEADER = _descriptor.Descriptor(
  name='BlockHeader',
  full_name='types.BlockHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='chainID', full_name='types.BlockHeader.chainID', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prevBlockHash', full_name='types.BlockHeader.prevBlockHash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blockNo', full_name='types.BlockHeader.blockNo', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='types.BlockHeader.timestamp', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blocksRootHash', full_name='types.BlockHeader.blocksRootHash', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='txsRootHash', full_name='types.BlockHeader.txsRootHash', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receiptsRootHash', full_name='types.BlockHeader.receiptsRootHash', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='confirms', full_name='types.BlockHeader.confirms', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pubKey', full_name='types.BlockHeader.pubKey', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='coinbaseAccount', full_name='types.BlockHeader.coinbaseAccount', index=9,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sign', full_name='types.BlockHeader.sign', index=10,
      number=11, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=119,
  serialized_end=352,
)


_BLOCKBODY = _descriptor.Descriptor(
  name='BlockBody',
  full_name='types.BlockBody',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='txs', full_name='types.BlockBody.txs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=354,
  serialized_end=389,
)


_TXLIST = _descriptor.Descriptor(
  name='TxList',
  full_name='types.TxList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='txs', full_name='types.TxList.txs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=391,
  serialized_end=423,
)


_TX = _descriptor.Descriptor(
  name='Tx',
  full_name='types.Tx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash', full_name='types.Tx.hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='body', full_name='types.Tx.body', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=425,
  serialized_end=472,
)


_TXBODY = _descriptor.Descriptor(
  name='TxBody',
  full_name='types.TxBody',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nonce', full_name='types.TxBody.nonce', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='account', full_name='types.TxBody.account', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='recipient', full_name='types.TxBody.recipient', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amount', full_name='types.TxBody.amount', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='payload', full_name='types.TxBody.payload', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gasLimit', full_name='types.TxBody.gasLimit', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gasPrice', full_name='types.TxBody.gasPrice', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='types.TxBody.type', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chainIdHash', full_name='types.TxBody.chainIdHash', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sign', full_name='types.TxBody.sign', index=9,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=475,
  serialized_end=667,
)


_TXIDX = _descriptor.Descriptor(
  name='TxIdx',
  full_name='types.TxIdx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='blockHash', full_name='types.TxIdx.blockHash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='idx', full_name='types.TxIdx.idx', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=669,
  serialized_end=708,
)


_TXINBLOCK = _descriptor.Descriptor(
  name='TxInBlock',
  full_name='types.TxInBlock',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='txIdx', full_name='types.TxInBlock.txIdx', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx', full_name='types.TxInBlock.tx', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=710,
  serialized_end=773,
)


_STATE = _descriptor.Descriptor(
  name='State',
  full_name='types.State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nonce', full_name='types.State.nonce', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='types.State.balance', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='codeHash', full_name='types.State.codeHash', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='storageRoot', full_name='types.State.storageRoot', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sqlRecoveryPoint', full_name='types.State.sqlRecoveryPoint', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=775,
  serialized_end=879,
)


_ACCOUNTPROOF = _descriptor.Descriptor(
  name='AccountProof',
  full_name='types.AccountProof',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='types.AccountProof.state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inclusion', full_name='types.AccountProof.inclusion', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='types.AccountProof.key', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proofKey', full_name='types.AccountProof.proofKey', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proofVal', full_name='types.AccountProof.proofVal', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bitmap', full_name='types.AccountProof.bitmap', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='types.AccountProof.height', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='auditPath', full_name='types.AccountProof.auditPath', index=7,
      number=8, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=882,
  serialized_end=1044,
)


_CONTRACTVARPROOF = _descriptor.Descriptor(
  name='ContractVarProof',
  full_name='types.ContractVarProof',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='types.ContractVarProof.value', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inclusion', full_name='types.ContractVarProof.inclusion', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proofKey', full_name='types.ContractVarProof.proofKey', index=2,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proofVal', full_name='types.ContractVarProof.proofVal', index=3,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bitmap', full_name='types.ContractVarProof.bitmap', index=4,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='types.ContractVarProof.height', index=5,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='auditPath', full_name='types.ContractVarProof.auditPath', index=6,
      number=8, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='types.ContractVarProof.key', index=7,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1047,
  serialized_end=1205,
)


_STATEQUERYPROOF = _descriptor.Descriptor(
  name='StateQueryProof',
  full_name='types.StateQueryProof',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='contractProof', full_name='types.StateQueryProof.contractProof', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='varProofs', full_name='types.StateQueryProof.varProofs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1207,
  serialized_end=1312,
)


_RECEIPT = _descriptor.Descriptor(
  name='Receipt',
  full_name='types.Receipt',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='contractAddress', full_name='types.Receipt.contractAddress', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='status', full_name='types.Receipt.status', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ret', full_name='types.Receipt.ret', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='txHash', full_name='types.Receipt.txHash', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='feeUsed', full_name='types.Receipt.feeUsed', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cumulativeFeeUsed', full_name='types.Receipt.cumulativeFeeUsed', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bloom', full_name='types.Receipt.bloom', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='events', full_name='types.Receipt.events', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blockNo', full_name='types.Receipt.blockNo', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blockHash', full_name='types.Receipt.blockHash', index=9,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='txIndex', full_name='types.Receipt.txIndex', index=10,
      number=11, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='from', full_name='types.Receipt.from', index=11,
      number=12, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='to', full_name='types.Receipt.to', index=12,
      number=13, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1315,
  serialized_end=1562,
)


_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='types.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='contractAddress', full_name='types.Event.contractAddress', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eventName', full_name='types.Event.eventName', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jsonArgs', full_name='types.Event.jsonArgs', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eventIdx', full_name='types.Event.eventIdx', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='txHash', full_name='types.Event.txHash', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blockHash', full_name='types.Event.blockHash', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blockNo', full_name='types.Event.blockNo', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='txIndex', full_name='types.Event.txIndex', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1565,
  serialized_end=1721,
)


_FNARGUMENT = _descriptor.Descriptor(
  name='FnArgument',
  full_name='types.FnArgument',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='types.FnArgument.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1723,
  serialized_end=1749,
)


_FUNCTION = _descriptor.Descriptor(
  name='Function',
  full_name='types.Function',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='types.Function.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='types.Function.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='payable', full_name='types.Function.payable', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='view', full_name='types.Function.view', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1751,
  serialized_end=1844,
)


_STATEVAR = _descriptor.Descriptor(
  name='StateVar',
  full_name='types.StateVar',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='types.StateVar.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='types.StateVar.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='len', full_name='types.StateVar.len', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1846,
  serialized_end=1897,
)


_ABI = _descriptor.Descriptor(
  name='ABI',
  full_name='types.ABI',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='types.ABI.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='language', full_name='types.ABI.language', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='functions', full_name='types.ABI.functions', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state_variables', full_name='types.ABI.state_variables', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1899,
  serialized_end=2017,
)


_QUERY = _descriptor.Descriptor(
  name='Query',
  full_name='types.Query',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='contractAddress', full_name='types.Query.contractAddress', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='queryinfo', full_name='types.Query.queryinfo', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2019,
  serialized_end=2070,
)


_STATEQUERY = _descriptor.Descriptor(
  name='StateQuery',
  full_name='types.StateQuery',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='contractAddress', full_name='types.StateQuery.contractAddress', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='root', full_name='types.StateQuery.root', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='compressed', full_name='types.StateQuery.compressed', index=2,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='storageKeys', full_name='types.StateQuery.storageKeys', index=3,
      number=5, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2072,
  serialized_end=2170,
)


_FILTERINFO = _descriptor.Descriptor(
  name='FilterInfo',
  full_name='types.FilterInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='contractAddress', full_name='types.FilterInfo.contractAddress', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eventName', full_name='types.FilterInfo.eventName', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blockfrom', full_name='types.FilterInfo.blockfrom', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blockto', full_name='types.FilterInfo.blockto', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='desc', full_name='types.FilterInfo.desc', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='argFilter', full_name='types.FilterInfo.argFilter', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='recentBlockCnt', full_name='types.FilterInfo.recentBlockCnt', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2173,
  serialized_end=2322,
)

_BLOCK.fields_by_name['header'].message_type = _BLOCKHEADER
_BLOCK.fields_by_name['body'].message_type = _BLOCKBODY
_BLOCKBODY.fields_by_name['txs'].message_type = _TX
_TXLIST.fields_by_name['txs'].message_type = _TX
_TX.fields_by_name['body'].message_type = _TXBODY
_TXBODY.fields_by_name['type'].enum_type = _TXTYPE
_TXINBLOCK.fields_by_name['txIdx'].message_type = _TXIDX
_TXINBLOCK.fields_by_name['tx'].message_type = _TX
_ACCOUNTPROOF.fields_by_name['state'].message_type = _STATE
_STATEQUERYPROOF.fields_by_name['contractProof'].message_type = _ACCOUNTPROOF
_STATEQUERYPROOF.fields_by_name['varProofs'].message_type = _CONTRACTVARPROOF
_RECEIPT.fields_by_name['events'].message_type = _EVENT
_FUNCTION.fields_by_name['arguments'].message_type = _FNARGUMENT
_ABI.fields_by_name['functions'].message_type = _FUNCTION
_ABI.fields_by_name['state_variables'].message_type = _STATEVAR
DESCRIPTOR.message_types_by_name['Block'] = _BLOCK
DESCRIPTOR.message_types_by_name['BlockHeader'] = _BLOCKHEADER
DESCRIPTOR.message_types_by_name['BlockBody'] = _BLOCKBODY
DESCRIPTOR.message_types_by_name['TxList'] = _TXLIST
DESCRIPTOR.message_types_by_name['Tx'] = _TX
DESCRIPTOR.message_types_by_name['TxBody'] = _TXBODY
DESCRIPTOR.message_types_by_name['TxIdx'] = _TXIDX
DESCRIPTOR.message_types_by_name['TxInBlock'] = _TXINBLOCK
DESCRIPTOR.message_types_by_name['State'] = _STATE
DESCRIPTOR.message_types_by_name['AccountProof'] = _ACCOUNTPROOF
DESCRIPTOR.message_types_by_name['ContractVarProof'] = _CONTRACTVARPROOF
DESCRIPTOR.message_types_by_name['StateQueryProof'] = _STATEQUERYPROOF
DESCRIPTOR.message_types_by_name['Receipt'] = _RECEIPT
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['FnArgument'] = _FNARGUMENT
DESCRIPTOR.message_types_by_name['Function'] = _FUNCTION
DESCRIPTOR.message_types_by_name['StateVar'] = _STATEVAR
DESCRIPTOR.message_types_by_name['ABI'] = _ABI
DESCRIPTOR.message_types_by_name['Query'] = _QUERY
DESCRIPTOR.message_types_by_name['StateQuery'] = _STATEQUERY
DESCRIPTOR.message_types_by_name['FilterInfo'] = _FILTERINFO
DESCRIPTOR.enum_types_by_name['TxType'] = _TXTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Block = _reflection.GeneratedProtocolMessageType('Block', (_message.Message,), dict(
  DESCRIPTOR = _BLOCK,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.Block)
  ))
_sym_db.RegisterMessage(Block)

BlockHeader = _reflection.GeneratedProtocolMessageType('BlockHeader', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKHEADER,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.BlockHeader)
  ))
_sym_db.RegisterMessage(BlockHeader)

BlockBody = _reflection.GeneratedProtocolMessageType('BlockBody', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKBODY,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.BlockBody)
  ))
_sym_db.RegisterMessage(BlockBody)

TxList = _reflection.GeneratedProtocolMessageType('TxList', (_message.Message,), dict(
  DESCRIPTOR = _TXLIST,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.TxList)
  ))
_sym_db.RegisterMessage(TxList)

Tx = _reflection.GeneratedProtocolMessageType('Tx', (_message.Message,), dict(
  DESCRIPTOR = _TX,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.Tx)
  ))
_sym_db.RegisterMessage(Tx)

TxBody = _reflection.GeneratedProtocolMessageType('TxBody', (_message.Message,), dict(
  DESCRIPTOR = _TXBODY,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.TxBody)
  ))
_sym_db.RegisterMessage(TxBody)

TxIdx = _reflection.GeneratedProtocolMessageType('TxIdx', (_message.Message,), dict(
  DESCRIPTOR = _TXIDX,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.TxIdx)
  ))
_sym_db.RegisterMessage(TxIdx)

TxInBlock = _reflection.GeneratedProtocolMessageType('TxInBlock', (_message.Message,), dict(
  DESCRIPTOR = _TXINBLOCK,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.TxInBlock)
  ))
_sym_db.RegisterMessage(TxInBlock)

State = _reflection.GeneratedProtocolMessageType('State', (_message.Message,), dict(
  DESCRIPTOR = _STATE,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.State)
  ))
_sym_db.RegisterMessage(State)

AccountProof = _reflection.GeneratedProtocolMessageType('AccountProof', (_message.Message,), dict(
  DESCRIPTOR = _ACCOUNTPROOF,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.AccountProof)
  ))
_sym_db.RegisterMessage(AccountProof)

ContractVarProof = _reflection.GeneratedProtocolMessageType('ContractVarProof', (_message.Message,), dict(
  DESCRIPTOR = _CONTRACTVARPROOF,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.ContractVarProof)
  ))
_sym_db.RegisterMessage(ContractVarProof)

StateQueryProof = _reflection.GeneratedProtocolMessageType('StateQueryProof', (_message.Message,), dict(
  DESCRIPTOR = _STATEQUERYPROOF,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.StateQueryProof)
  ))
_sym_db.RegisterMessage(StateQueryProof)

Receipt = _reflection.GeneratedProtocolMessageType('Receipt', (_message.Message,), dict(
  DESCRIPTOR = _RECEIPT,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.Receipt)
  ))
_sym_db.RegisterMessage(Receipt)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), dict(
  DESCRIPTOR = _EVENT,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.Event)
  ))
_sym_db.RegisterMessage(Event)

FnArgument = _reflection.GeneratedProtocolMessageType('FnArgument', (_message.Message,), dict(
  DESCRIPTOR = _FNARGUMENT,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.FnArgument)
  ))
_sym_db.RegisterMessage(FnArgument)

Function = _reflection.GeneratedProtocolMessageType('Function', (_message.Message,), dict(
  DESCRIPTOR = _FUNCTION,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.Function)
  ))
_sym_db.RegisterMessage(Function)

StateVar = _reflection.GeneratedProtocolMessageType('StateVar', (_message.Message,), dict(
  DESCRIPTOR = _STATEVAR,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.StateVar)
  ))
_sym_db.RegisterMessage(StateVar)

ABI = _reflection.GeneratedProtocolMessageType('ABI', (_message.Message,), dict(
  DESCRIPTOR = _ABI,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.ABI)
  ))
_sym_db.RegisterMessage(ABI)

Query = _reflection.GeneratedProtocolMessageType('Query', (_message.Message,), dict(
  DESCRIPTOR = _QUERY,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.Query)
  ))
_sym_db.RegisterMessage(Query)

StateQuery = _reflection.GeneratedProtocolMessageType('StateQuery', (_message.Message,), dict(
  DESCRIPTOR = _STATEQUERY,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.StateQuery)
  ))
_sym_db.RegisterMessage(StateQuery)

FilterInfo = _reflection.GeneratedProtocolMessageType('FilterInfo', (_message.Message,), dict(
  DESCRIPTOR = _FILTERINFO,
  __module__ = 'blockchain_pb2'
  # @@protoc_insertion_point(class_scope:types.FilterInfo)
  ))
_sym_db.RegisterMessage(FilterInfo)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

import json
import datetime

from .address import Address
from .block_hash import BlockHash
from .transaction import Transaction
from ..utils.encoding import encode_b58, encode_b58_check
from ..utils.converter import get_hash


class Block:
    def __init__(self, hash_value=None, height=None, grpc_block=None,
                 grpc_block_header=None, tx_cnt=0, size=0):
        if grpc_block is not None:
            self._map_grpc_block(grpc_block)
        elif grpc_block_header is not None:
            if hash_value is None:
                raise ValueError("Cannot set without a block hash value")

            # block hash
            if type(hash_value) is not BlockHash:
                hash_value = BlockHash(hash_value)
            self.__hash = hash_value
            # header
            self._map_grpc_block_header(grpc_block_header)
            # body
            self.__tx_list = []
            self.__tx_cnt = tx_cnt
            self.__size = size
        else:
            if hash_value is None:
                raise ValueError("Cannot set without a block hash value")

            # block hash
            if type(hash_value) is not BlockHash:
                hash_value = BlockHash(hash_value)
            self.__hash = hash_value
            # header
            self.__chain_id = None
            self.__prev_block = None
            self.__height = height
            self.__timestamp = None
            self.__blocks_root_hash = None
            self.__txs_root_hash = None
            self.__receipts_root_hash = None
            self.__confirms = None
            self.__public_key = None
            self.__sign = None
            self.__coinbase_account = None
            # body
            self.__tx_list = []
            self.__tx_cnt = tx_cnt
            self.__size = size

    def _map_grpc_block_header(self, header):
        # header
        self.__chain_id = header.chainID
        if header.blockNo > 0:
            self.__prev_block = Block(hash_value=header.prevBlockHash,
                                      height=header.blockNo - 1)
        else:
            self.__prev_block = None
        self.__height = header.blockNo
        self.__timestamp = header.timestamp
        self.__blocks_root_hash = header.blocksRootHash
        self.__txs_root_hash = header.txsRootHash
        self.__receipts_root_hash = header.receiptsRootHash
        self.__confirms = header.confirms
        self.__public_key = header.pubKey
        self.__sign = header.sign
        self.__coinbase_account = header.coinbaseAccount

    def _map_grpc_block(self, v):
        self.__hash = BlockHash(v.hash)
        # header
        self._map_grpc_block_header(v.header)
        # body
        self.__tx_list = []
        for i, tx in enumerate(v.body.txs):
            from_address = Address(None, empty=True)
            from_address.value = tx.body.account
            to_address = Address(None, empty=True)
            to_address.value = tx.body.recipient

            block_tx = Transaction(read_only=True, tx_hash=tx.hash,
                                   nonce=tx.body.nonce,
                                   from_address=from_address,
                                   to_address=to_address,
                                   amount=tx.body.amount,
                                   payload=tx.body.payload,
                                   fee_price=tx.body.gasPrice,
                                   fee_limit=tx.body.gasLimit,
                                   tx_sign=tx.body.sign, tx_type=tx.body.type,
                                   block=self, index_in_block=i,
                                   is_in_mempool=False)
            self.__tx_list.append(block_tx)
        self.__tx_cnt = len(self.__tx_list)
        self.__size = v.ByteSize()

    @property
    def hash(self):
        return self.__hash

    @property
    def chain_id(self):
        return self.__chain_id

    @property
    def chain_id_hash(self):
        if self.__chain_id is None:
            return None
        return get_hash(self.__chain_id, no_rand=True, no_encode=True)

    @property
    def chain_id_hash_b58(self):
        if self.__chain_id is None:
            return None
        return get_hash(self.__chain_id, no_rand=True, no_encode=False)

    @property
    def height(self):
        return self.__height

    @property
    def prev(self):
        return self.__prev_block

    @property
    def block_no(self):
        return self.height

    @property
    def timestamp(self):
        return self.__timestamp

    @property
    def datetimestamp(self):
        if self.__timestamp is None:
            return None

        return str(datetime.datetime.fromtimestamp(self.timestamp // 1000000000))

    @property
    def blocks_root_hash(self):
        return self.__blocks_root_hash

    @property
    def txs_root_hash(self):
        return self.__txs_root_hash

    @property
    def receipts_root_hash(self):
        return self.__receipts_root_hash

    @property
    def confirms(self):
        return self.__confirms

    @property
    def public_key(self):
        return self.__public_key

    @property
    def sign(self):
        return self.__sign

    @property
    def coinbase_account(self):
        return self.__coinbase_account

    @property
    def tx_list(self):
        return self.__tx_list

    @property
    def num_of_tx(self):
        return self.__tx_cnt

    @property
    def size(self):
        return self.__size

    def get_tx(self, index):
        return self.__tx_list[index]

    def json(self, header_only=False):
        body_json = {
            "hash": str(self.hash),
            "header": {
                "chain_id": self.chain_id_hash_b58,
                "previous_block_hash": str(self.prev.hash) if self.prev is not None else None,
                "block_no": self.block_no,
                "timestamp": self.timestamp,
                "datetimestamp": self.datetimestamp,
                "blocks_root_hash": encode_b58(self.blocks_root_hash),
                "txs_root_hash": encode_b58(self.txs_root_hash),
                "receipts_root_hash": encode_b58(self.receipts_root_hash),
                "confirms": self.confirms,
                "pub_key": encode_b58(self.public_key) if self.public_key is not None else None,
                "sign": encode_b58(self.sign),
                "coinbase_account": encode_b58_check(self.coinbase_account),
                "tx_count": self.__tx_cnt,
                "size": self.__size
            },
        }

        if not header_only:
            tx_list = []
            for tx in self.tx_list:
                tx_list.append(tx.json())

            body_json["body"] = {
                "tx_list": tx_list,
            }

        return body_json

    def __str__(self):
        return json.dumps(self.json(), indent=2)



# -*- coding: utf-8 -*-

"""Main module."""

import hashlib
import json
import time

from . import account as acc
from . import comm

from .obj import transaction
from .obj import address as addr
from .obj import block_hash as bh
from .obj import peer as pr
from .obj import tx_hash as th
from .obj.block import Block
from .obj.block_stream import BlockStream
from .obj.blockchain_info import BlockchainInfo
from .obj.blockchain_status import BlockchainStatus
from .obj.call_info import CallInfo
from .obj.change_conf_info import ChangeConfInfo
from .obj.consensus_info import ConsensusInfo
from .obj.event import Event
from .obj.event_stream import EventStream
from .obj.name_info import NameInfo
from .obj.node_info import NodeInfo
from .obj.sc_state import SCState, SCStateVar
from .obj.transaction import Transaction, TxType
from .obj.tx_hash import TxHash
from .obj.tx_result import TxResult
from .obj.var_proof import VarProofs
from .obj.abi import Abi
from .obj.enterprise_config import EnterpriseConfig

from .errors.exception import CommunicationException
from .errors.general_exception import GeneralException

from .status.commit_status import CommitStatus

from .utils.encoding import decode_address, \
    encode_private_key, decode_private_key, decode_root, \
    encode_tx_hash, decode_tx_hash


class Aergo:
    def __init__(self):
        self.__account = None
        self.__comm = None

    @property
    def account(self):
        """
        Returns the account object.
        :return:
        """
        return self.__account

    @account.setter
    def account(self, a):
        self.__account = a

    def new_account(self, private_key=None, skip_state=False):
        self.__account = acc.Account(private_key=private_key)
        if not skip_state:
            self.get_account()
        return self.__account

    # TODO how about making account_state class,
    #       or how about returning account and change method name
    def get_account(self, account=None, address=None, proof=False, root=b'', compressed=True):
        """
        Return account information
        :param address:
        :param proof:
        :param root:
        :param compressed:
        :return:
        """
        if self.__comm is None:
            return None

        if account is None and address is None:
            # self account
            ret_account = self.__account
            req_address = bytes(self.__account.address)
        elif account is not None:
            ret_account = account
            req_address = bytes(account.address)
        else:
            if address is None:
                raise GeneralException("Fail to get an account info. No designated account address.")
            else:
                ret_account = acc.Account(empty=True)
                if isinstance(address, str):
                    req_address = addr.Address.decode(address)
                elif type(address) is addr.Address:
                    req_address = bytes(address)
                else:
                    req_address = address
                ret_account.address = req_address

        if proof:
            if isinstance(root, str) and len(root) != 0:
                root = decode_root(root)
            try:
                state_proof = self.__comm.get_account_state_proof(req_address,
                                                                  root,
                                                                  compressed)
            except Exception as e:
                raise CommunicationException(e) from e

            ret_account.state = state_proof.state
            ret_account.state_proof = state_proof
        else:
            try:
                state = self.__comm.get_account_state(req_address)
            except Exception as e:
                raise CommunicationException(e) from e

            ret_account.state = state
            ret_account.state_proof = None

        return ret_account

    def connect(self, target):
        """
        Connect to the gRPC server running on port `target` e.g.
        target="localhost:7845".
        :param target:
        :return:
        """
        if target is None:
            raise ValueError('need target value')

        self.__comm = comm.Comm(target)
        try:
            self.__comm.connect()
        except Exception as e:
            raise CommunicationException(e) from e
        try:
            status = self.__comm.get_blockchain_status()
        except Exception as e:
            raise CommunicationException(e) from e
        self.chain_id = status.best_chain_id_hash

    def disconnect(self):
        """
        Disconnect from the gRPC server.
        """
        if self.__comm is not None:
            try:
                self.__comm.disconnect()
            except Exception as e:
                raise CommunicationException(e) from e

    def get_chain_info(self, with_consensus_info=True):
        """
        Returns the blockchain info
        :return:
        """
        if self.__comm is None:
            return None

        try:
            chain_info = self.__comm.get_chain_info()
            if with_consensus_info:
                consensus_info = self.__comm.get_consensus_info()
            else:
                consensus_info = None
        except Exception as e:
            raise CommunicationException(e) from e

        return BlockchainInfo(chain_info, consensus_info)

    def get_consensus_info(self):
        """
        Returns the consensus information
        :return:
        """
        if self.__comm is None:
            return None

        try:
            info = self.__comm.get_consensus_info()
        except Exception as e:
            raise CommunicationException(e) from e

        return ConsensusInfo(info)

    def get_status(self):
        """
        Returns the blockchain status
        :return:
        """
        if self.__comm is None:
            return None

        try:
            status = self.__comm.get_blockchain_status()
        except Exception as e:
            raise CommunicationException(e) from e

        return BlockchainStatus(status)

    def receive_block_meta_stream(self):
        """
        Returns the iterable block stream
        :return:
        """
        if self.__comm is None:
            return None

        try:
            stream = self.__comm.receive_block_meta_stream()
        except Exception as e:
            raise CommunicationException(e) from e

        return BlockStream(stream)

    def receive_block_stream(self):
        """
        Returns the iterable block stream
        :return:
        """
        if self.__comm is None:
            return None

        try:
            stream = self.__comm.receive_block_stream()
        except Exception as e:
            raise CommunicationException(e) from e

        return BlockStream(stream)

    def get_block_metas(self, block_hash=None, block_height=-1, list_size=20,
                        offset=0, is_asc_order=False):
        """
        Returns the list of metadata of queried blocks.
        :param block_hash:
        :param block_height:
        :param list_size: maximum number of results
        :param offset: the start point to search until the block_hash or block_height
        :param is_asc_order:
        :return:
        """
        if self.__comm is None:
            return None

        if block_hash is None and block_height < 0:
            raise ValueError("Please insert a block hash or height")

        try:
            bms = self.__comm.get_block_metas(block_hash=block_hash,
                                              block_height=block_height,
                                              list_size=list_size,
                                              offset=offset,
                                              is_asc_order=is_asc_order)
            block_headers = []
            for bm in bms.blocks:
                block_headers.append(Block(hash_value=bm.hash,
                                           grpc_block_header=bm.header,
                                           tx_cnt=bm.txcount,
                                           size=bm.size))
        except Exception as e:
            raise CommunicationException(e) from e

        return block_headers

    def receive_event_stream(self, sc_address, event_name, start_block_no=0,
                             end_block_no=0, with_desc=False, arg_filter=None,
                             recent_block_cnt=0):
        if self.__comm is None:
            return None

        if isinstance(sc_address, str):
            # TODO exception handling: raise ValueError("Invalid checksum")
            sc_address = addr.Address.decode(sc_address)
        elif isinstance(sc_address, TxHash):
            sc_address = bytes(sc_address)

        if arg_filter:
            if isinstance(arg_filter, (dict, list, tuple)):
                arg_filter = json.dumps(arg_filter)
            arg_filter = arg_filter.encode('utf-8')

        try:
            es = self.__comm.receive_event_stream(sc_address=sc_address,
                                                  event_name=event_name,
                                                  start_block_no=start_block_no,
                                                  end_block_no=end_block_no,
                                                  with_desc=with_desc,
                                                  arg_filter=arg_filter,
                                                  recent_block_cnt=recent_block_cnt)
        except Exception as e:
            raise CommunicationException(e) from e

        return EventStream(es)

    def get_events(self, sc_address, event_name, start_block_no=-1,
                   end_block_no=-1, with_desc=False, arg_filter=None,
                   recent_block_cnt=0):
        if self.__comm is None:
            return None

        if isinstance(sc_address, str):
            # TODO exception handling: raise ValueError("Invalid checksum")
            sc_address = addr.Address.decode(sc_address)
        elif isinstance(sc_address, TxHash):
            sc_address = bytes(sc_address)

        if arg_filter:
            if isinstance(arg_filter, (dict, list, tuple)):
                arg_filter = json.dumps(arg_filter)
            arg_filter = arg_filter.encode('utf-8')

        # max range = 10000
        if start_block_no < 0 and end_block_no < 0:
            recent_block_cnt = 10000
        elif start_block_no < 0:
            start_block_no = end_block_no - 10000
            if start_block_no < 0:
                start_block_no = 0
        elif end_block_no < 0:
            end_block_no = start_block_no + 10000

        try:
            result = self.__comm.get_events(sc_address=sc_address,
                                            event_name=event_name,
                                            start_block_no=start_block_no,
                                            end_block_no=end_block_no,
                                            with_desc=with_desc,
                                            arg_filter=arg_filter,
                                            recent_block_cnt=recent_block_cnt)
            event_list = []
            for e in result.events:
                event_list.append(Event(e))
        except Exception as e:
            raise CommunicationException(e) from e

        return event_list

    def get_block_headers(self, block_hash=None, block_height=-1, list_size=20,
                          offset=0, is_asc_order=False):
        """
        Returns the list of blocks.
        :param block_hash:
        :param block_height:
        :param list_size: maximum number of results
        :param offset: the start point to search until the block_hash or block_height
        :param is_asc_order:
        :return:
        """
        if self.__comm is None:
            return None

        if block_hash is None and block_height < 0:
            raise ValueError("Please insert a block hash or height")

        try:
            bhs = self.__comm.get_block_headers(block_hash=block_hash,
                                                block_height=block_height,
                                                list_size=list_size,
                                                offset=offset,
                                                is_asc_order=is_asc_order)
            block_headers = []
            for b in bhs.blocks:
                block_headers.append(Block(grpc_block=b))
        except Exception as e:
            raise CommunicationException(e) from e

        return block_headers

    def get_blockchain_status(self):
        """
        Returns the highest block hash and block height so far.
        :return:
        """
        status = self.get_status()
        return status.best_block_hash, status.best_block_height

    def get_block(self, block_hash=None, block_height=-1):
        """
        Returns block information for `block_hash` or `block_height`.
        :param block_hash:
        :param block_height:
        :return:
        """
        if self.__comm is None:
            return None

        if block_height >= 0:
            query = block_height.to_bytes(8, byteorder='little')
        else:
            if type(block_hash) is not bh.BlockHash:
                block_hash = bh.BlockHash(block_hash)
            query = block_hash.value

        try:
            result = self.__comm.get_block(query)
        except Exception as e:
            raise CommunicationException(e) from e

        b = Block(grpc_block=result)
        return b

    def get_block_meta(self, block_hash=None, block_height=-1):
        """
        Returns block metadata for `block_hash` or `block_height`.
        :param block_hash:
        :param block_height:
        :return:
        """
        if self.__comm is None:
            return None

        if block_height >= 0:
            query = block_height.to_bytes(8, byteorder='little')
        else:
            if type(block_hash) is not bh.BlockHash:
                block_hash = bh.BlockHash(block_hash)
            query = block_hash.value

        try:
            bm = self.__comm.get_block_meta(query)
        except Exception as e:
            raise CommunicationException(e) from e

        b = Block(hash_value=bm.hash,
                  grpc_block_header=bm.header,
                  tx_cnt=bm.txcount,
                  size=bm.size)
        return b

    def get_node_accounts(self, skip_state=False):
        """
        Returns a list of all node accounts.
        :return:
        """

        try:
            result = self.__comm.get_accounts()
        except Exception as e:
            raise CommunicationException(e) from e

        accounts = []
        for a in result.accounts:
            if skip_state:
                account = acc.Account(empty=True)
                account.address = a.address
            else:
                account = self.get_account(address=a.address)
            accounts.append(account)

        return accounts

    def get_peers(self):
        """
        Returns a list of peers.
        :return:
        """
        try:
            result = self.__comm.get_peers()
        except Exception as e:
            raise CommunicationException(e) from e

        peers = []
        for i, p in enumerate(result.peers):
            peer = pr.Peer()
            peer.info = p
            peers.append(peer)

        return peers

    def get_node_info(self, keys=None):
        """
        Returns the consensus information
        :return:
        """
        if self.__comm is None:
            return None

        try:
            info = self.__comm.get_node_info(keys)
        except Exception as e:
            raise CommunicationException(e) from e

        return NodeInfo(info)

    def get_node_state(self, timeout=1):
        """
        Returns information about the node state.
        :return:
        """
        try:
            result = self.__comm.get_node_state(timeout)
        except Exception as e:
            raise CommunicationException(e) from e

        json_txt = result.value.decode('utf8').replace("'", '"')
        return json.loads(json_txt)

    def get_tx(self, tx_hash, mempool_only=False, skip_block=False):
        """
        Returns info on transaction with hash `tx_hash`.
        :param tx_hash:
        :return:
        """
        if isinstance(tx_hash, str):
            tx_hash = decode_tx_hash(tx_hash)
        elif type(tx_hash) == TxHash:
            tx_hash = bytes(tx_hash)

        result_tx = None
        result_tx_block_hash = None
        result_tx_index = None
        result_tx_is_in_mempool = False

        try:
            result_tx = self.__comm.get_tx(tx_hash)
            result_tx_block_hash = None
            result_tx_index = None
            result_tx_is_in_mempool = True
            get_result = True
        except Exception as e:
            if mempool_only:
                raise CommunicationException(e) from e
            else:
                get_result = False

        if not get_result:
            try:
                result = self.__comm.get_block_tx(tx_hash)
                result_tx = result.tx
                result_tx_block_hash = result.txIdx.blockHash
                result_tx_index = result.txIdx.idx
                result_tx_is_in_mempool = False
            except Exception as e:
                raise CommunicationException(e) from e

        if result_tx_block_hash is not None:
            if skip_block:
                result_tx_block = Block(hash_value=result_tx_block_hash,
                                        height=None)
            else:
                result_tx_block = self.get_block(block_hash=result_tx_block_hash)
        else:
            result_tx_block = None

        from_address = addr.Address(None, empty=True)
        from_address.value = result_tx.body.account
        to_address = addr.Address(None, empty=True)
        to_address.value = result_tx.body.recipient

        tx = Transaction(read_only=True,
                         tx_hash=result_tx.hash,
                         nonce=result_tx.body.nonce,
                         from_address=from_address,
                         to_address=to_address,
                         amount=result_tx.body.amount,
                         payload=result_tx.body.payload,
                         fee_price=result_tx.body.gasPrice,
                         fee_limit=result_tx.body.gasLimit,
                         tx_sign=result_tx.body.sign,
                         tx_type=result_tx.body.type,
                         chain_id=result_tx.body.chainIdHash,
                         block=result_tx_block, index_in_block=result_tx_index,
                         is_in_mempool=result_tx_is_in_mempool)

        return tx

    def lock_account(self, address, passphrase):
        """
        Locks the account with address `address` with the passphrase
        `passphrase`.
        :param address:
        :param passphrase:
        :return:
        """
        try:
            result = self.__comm.lock_account(address, passphrase)
        except Exception as e:
            raise CommunicationException(e) from e

        return result

    def unlock_account(self, address, passphrase):
        """
        Unlocks the account with address `address` with the passphrase
        `passphrase`.
        :param address:
        :param passphrase:
        :return:
        """
        try:
            result = self.__comm.unlock_account(address=address,
                                                passphrase=passphrase)
        except Exception as e:
            raise CommunicationException(e) from e

        return result

    def generate_tx(self, to_address, nonce, amount, fee_limit=0, fee_price=0,
                    payload=None, tx_type=TxType.NORMAL):
        if to_address is not None:
            address = addr.Address(None, empty=True)
            address.value = to_address
            to_address = address

        tx = transaction.Transaction(tx_type=tx_type,
                                     from_address=self.__account.address,
                                     to_address=to_address,
                                     nonce=nonce, amount=amount,
                                     fee_limit=fee_limit, fee_price=fee_price,
                                     payload=payload, chain_id=self.chain_id)
        tx.sign = self.__account.sign_msg_hash(tx.calculate_hash(including_sign=False))
        return tx

    def send_payload(self, amount, payload, to_address=None, retry_nonce=0,
                     tx_type=TxType.NORMAL):
        if self.__comm is None:
            return None, None

        if isinstance(to_address, str):
            address_type = addr.check_name_address(to_address)
            if address_type > 0:
                to_address = to_address.encode()
                if 2 == address_type:
                    tx_type = TxType.GOVERNANCE
            else:
                try:
                    to_address = decode_address(to_address)
                except Exception as e:
                    raise ValueError("Invalid receiver address: {}".format(e))
        elif isinstance(to_address, addr.GovernanceTxAddress):
            if addr.check_name_address(to_address.value):
                to_address = to_address.value.encode()
                tx_type = TxType.GOVERNANCE

        nonce = self.__account.nonce + 1
        tx = self.generate_tx(to_address=to_address,
                              nonce=nonce, amount=amount,
                              fee_limit=0, fee_price=0,
                              payload=payload, tx_type=tx_type)
        signed_tx, result = self.send_tx(signed_tx=tx)

        if result.status == CommitStatus.TX_OK:
            self.__account.nonce = nonce
        elif result.status == CommitStatus.TX_HAS_SAME_NONCE:
            while retry_nonce > 0:
                retry_nonce -= 1

                nonce += 1
                tx = self.generate_tx(to_address=to_address,
                                      nonce=nonce, amount=amount,
                                      fee_limit=0, fee_price=0,
                                      payload=payload, tx_type=tx_type)
                signed_tx, result = self.send_tx(signed_tx=tx)

                es = result.status
                if es == CommitStatus.TX_OK:
                    self.__account.nonce = nonce
                    break
                elif es != CommitStatus.TX_HAS_SAME_NONCE:
                    break

        return signed_tx, result

    def send_unsigned_tx(self, unsigned_tx):
        """
        Sends the unsigned transaction.
        The unsigned transaction will be signed by the account
        which is stored in the connected node.
        :param unsigned_tx:
        :return:
        """
        try:
            result = self.__comm.send_tx(unsigned_tx)
        except Exception as e:
            raise CommunicationException(e) from e

        return result

    def send_tx(self, signed_tx):
        """
        Send a signed transaction.
        This transaction will push to the memory pool after verifying.
        :param signed_tx:
        :return:
        """
        signed_txs, results = self.batch_tx(signed_txs=[signed_tx])
        return signed_txs[0], results[0]

    def batch_tx(self, signed_txs):
        """
        Send a set of signed transactions simultaneously.
        These transactions will push to the memory pool after verifying.
        :param signed_txs:
        :return:
        """
        try:
            result_list = self.__comm.commit_txs(signed_txs)
        except Exception as e:
            raise CommunicationException(e) from e

        results = []
        for i, r in enumerate(result_list.results):
            tx_result = TxResult(tx=signed_txs[i], result=result_list.results[i])
            if tx_result.status == CommitStatus.TX_OK:
                self.__account.nonce += 1
            results.append(tx_result)
        return signed_txs, results

    def import_account(self, exported_data, password, skip_state=False, skip_self=False):
        if exported_data is None or 0 == len(exported_data):
            # TODO unit test + exception handling
            assert 1 == 0

        if password is None or 0 == len(password):
            # TODO unit test + exception handling
            assert 1 == 0

        if isinstance(exported_data, str):
            exported_data = decode_private_key(exported_data)

        if isinstance(password, bytes):
            password = password.decode('utf-8')

        account = acc.Account.decrypt_account(exported_data, password)
        if not skip_self:
            self.__account = account
            if not skip_state:
                self.get_account()
        else:
            if not skip_state:
                self.get_account(account=account)
        return account

    def export_account(self, password, account=None):
        if account is None:
            account = self.__account

        enc_acc = acc.Account.encrypt_account(account, password)
        return encode_private_key(enc_acc)

    def get_tx_result(self, tx_hash):
        if self.__comm is None:
            return None

        if isinstance(tx_hash, str):
            tx_hash = decode_tx_hash(tx_hash)
        elif type(tx_hash) is th.TxHash:
            tx_hash = bytes(tx_hash)

        try:
            result = self.__comm.get_receipt(tx_hash)
            tx_result = TxResult(result=result)
            tx_result.tx_id = encode_tx_hash(tx_hash)
        except Exception as e:
            raise CommunicationException(e) from e

        return tx_result

    def wait_tx_result(self, tx_hash, timeout=30, tempo=0.2):
        if self.__comm is None:
            return None
        if isinstance(tx_hash, str):
            tx_hash = decode_tx_hash(tx_hash)
        elif type(tx_hash) is th.TxHash:
            tx_hash = bytes(tx_hash)

        for _ in range(int(timeout/tempo)+1):
            try:
                return self.get_tx_result(tx_hash)
            except CommunicationException as e:
                if e.error_details is None or e.error_details[:12] != "tx not found":
                    raise e
            time.sleep(tempo)
        return None
            
    def deploy_sc(self, payload, amount=0, args=None, retry_nonce=0,
                  redeploy=False):
        if isinstance(payload, str):
            payload = decode_address(payload)

        if args is not None and not isinstance(args, (list, tuple)):
            args = [args]

        payload_bytes = (len(payload) + 4).to_bytes(4, byteorder='little')
        payload_bytes += payload

        json_args = json.dumps(args, separators=(',', ':'))
        payload_bytes += json_args.encode('utf-8')

        if redeploy:
            tx_type = TxType.REDPLOY
        else:
            tx_type = TxType.NORMAL

        tx, result = self.send_payload(amount=amount, payload=payload_bytes,
                                       retry_nonce=retry_nonce, tx_type=tx_type)
        return tx, result

    def new_call_sc_tx(self, sc_address, func_name, amount=0, args=None,
                       nonce=None):
        tx_type = TxType.NORMAL
        if isinstance(sc_address, str):
            address_type = addr.check_name_address(sc_address)
            if address_type > 0:
                sc_address = sc_address.encode()
                if 2 == address_type:
                    tx_type = TxType.GOVERNANCE
            else:
                try:
                    sc_address = decode_address(sc_address)
                except Exception as e:
                    raise ValueError("Invalid smart contract address: {}".format(e))
        elif isinstance(sc_address, addr.GovernanceTxAddress):
            if addr.check_name_address(sc_address.value):
                sc_address = sc_address.value.encode()
                tx_type = TxType.GOVERNANCE

        if args is not None and not isinstance(args, (list, tuple)):
            args = [args]

        call_info = CallInfo(func_name, args).__dict__
        payload = json.dumps(call_info, separators=(',', ':')).encode('utf-8')

        if nonce is None:
            nonce = self.__account.nonce + 1

        return self.generate_tx(to_address=sc_address,
                                nonce=nonce, amount=amount,
                                fee_limit=0, fee_price=0,
                                payload=payload, tx_type=tx_type)

    def batch_call_sc(self, sc_txs):
        return self.batch_tx(sc_txs)

    def call_sc(self, sc_address, func_name, amount=0, args=None):
        sc_tx = self.new_call_sc_tx(sc_address=sc_address, func_name=func_name,
                                    amount=amount, args=args)
        sc_txs, results = self.batch_call_sc([sc_tx])
        return sc_txs[0], results[0]

    def query_sc(self, sc_address, func_name, args=None):
        if isinstance(sc_address, str):
            # TODO exception handling: raise ValueError("Invalid checksum")
            sc_address = addr.Address.decode(sc_address)

        if args is not None and not isinstance(args, (list, tuple)):
            args = [args]

        call_info = CallInfo(func_name, args).__dict__
        payload = json.dumps(call_info, separators=(',', ':')).encode('utf-8')

        try:
            result = self.__comm.query_contract(sc_address, payload)
        except Exception as e:
            raise CommunicationException(e) from e

        return result.value

    def query_sc_state(self, sc_address, storage_keys, root=b'',
                       compressed=True):
        """ query_sc_state returns a SCState object containing the contract
        state and variable state with their respective merkle proofs.
        """
        if isinstance(sc_address, str):
            # TODO exception handling: raise ValueError("Invalid checksum")
            sc_address = addr.Address.decode(sc_address)

        if isinstance(root, str) and len(root) != 0:
            root = decode_root(root)

        # convert SCStateVar objects to trie storage key strings
        trie_keys = []
        for key in storage_keys:
            if type(key) is bytes:
                trie_keys.append(hashlib.sha256(key).digest())
            elif type(key) is str:
                trie_keys.append(hashlib.sha256(key.encode('latin-1')).digest())
            elif type(key) is SCStateVar:
                trie_keys.append(hashlib.sha256(bytes(key)).digest())
            else:
                assert False, "Invalid key type provided, must be bytes, str or SCStateVar"

        try:
            result = self.__comm.query_contract_state(sc_address, trie_keys,
                                                      root, compressed)
        except Exception as e:
            raise CommunicationException(e) from e

        account = acc.Account(empty=True)
        account.state = result.contractProof.state
        account.state_proof = result.contractProof
        account.address = sc_address

        var_proofs = VarProofs(result.varProofs, trie_keys)

        return SCState(account=account, var_proofs=var_proofs)

    def get_conf_change_progress(self, block_height):
        """
        Returns the RAFT change config progress status after 'changeCluster' system contract
        :return:
        """
        if self.__comm is None:
            return None

        try:
            status = self.__comm.get_conf_change_progress(block_height)
        except Exception as e:
            raise CommunicationException(e) from e

        return ChangeConfInfo(status)

    def get_enterprise_config(self, key):
        if self.__comm is None:
            return None

        try:
            conf = self.__comm.get_enterprise_config(key)
        except Exception as e:
            raise CommunicationException(e) from e
        return EnterpriseConfig(conf)

    def get_name_info(self, name, block_height=-1):
        """
        Returns information of name which is designated by the system contract
        :param name:
        :param block_height:
        :return:
        """
        if self.__comm is None:
            return None

        if block_height < 0:
            # set current block height
            block_height = self.get_status().best_block_height

        try:
            info = self.__comm.get_name_info(name, block_height)
        except Exception as e:
            raise CommunicationException(e) from e

        return NameInfo(info)

    def get_abi(self, contract_addr=None, addr_bytes=None):
        """ Returns the abi of given contract address. """
        if self.__comm is None:
            return None
        if contract_addr is not None:
            addr_bytes = decode_address(contract_addr)
        try:
            abi = self.__comm.get_abi(addr_bytes)
        except Exception as e:
            raise CommunicationException(e) from e
        return Abi(abi)
